from sqlalchemy import create_engine
from pita.db_backend import *

def db_session(conn, new=False):
    if not hasattr(db_session, 'session') or not db_session.session:
        engine = create_engine(conn)
        engine.raw_connection().connection.text_factory = str
        db_session.engine = engine
        if new:
            Base.metadata.drop_all(db_session.engine)
        Base.metadata.create_all(engine)
        Base.metadata.bind = engine
        db_session.session = scoped_session(sessionmaker(bind=engine))
    elif new:
        db_session.session.commit()
        Base.metadata.drop_all(db_session.engine)
        Base.metadata.create_all(db_session.engine)
    
    return db_session.session



class Exon(object):
    def __init__(self, chrom, start, end, strand):
        if end < start:
            raise ValueError("exon end < start (%s:%s%s%s)" % (chrom, start, strand, end))
        self.chrom = chrom
        self.start = start
        self.end = end
        self.strand = strand
        self.evidence = []
        self.validated = False
        self.stats = {}
        self.seq = None

    def __nonzero__(self):
        return 1
    
    def __len__(self):
        return self.end - self.start

    def __repr__(self):
        return "%s:%s-%s" % (self.chrom, self.start, self.end)

    def add_evidence(self, ev):
        if ev not in self.evidence:
            self.evidence.append(ev)
    
    def overlap(self, exon, strand=True):
        if strand and self.strand != exon.strand:
            return 0
        if exon.start >= self.start and exon.start <= self.end:
            return self.end - exon.start
        if exon.end >= self.start and exon.end <= self.end:
            return exon.end - self.start
        return 0



import sys
import HTSeq
from pita.io import read_bed_transcripts
import numpy as np

def compare_annotation(models1, models2):

    ga = HTSeq.GenomicArrayOfSets("auto", stranded=False)
    l = {}
    result = {}
    for name, x, exons in models1:
        #exons[0][1] = exons[0][2] - 10
        #exons[-1][2] = exons[-1][1] + 10
        l[name] = 0.0
        for exon in exons:
            iv = HTSeq.GenomicInterval(*exon)
            ga[iv] += name
            l[name] += iv.end - iv.start

    for name, x, exons in models2:
        #exons[0][1] = exons[0][2] - 10
        #exons[-1][2] = exons[-1][1] + 10
        overlap = {}
        tlen = 0.0
        for exon in exons:
            test = HTSeq.GenomicInterval(*exon)
            
            for iv2, step_set in ga[test].steps():
                if len(step_set) > 0:
                    for n in step_set:
                        overlap[n] = overlap.setdefault(n, 0) + iv2.end - iv2.start
            tlen += exon[2] - exon[1]
        
        acs = []
        
        result[name] = {}
        for oname, length in overlap.items():
       
            #print "#", name, tlen, oname, l[oname], length
            #print name, oname 
            sn = length / l[oname]
            sp = length / tlen
            ac = (sn + sp) / 2
            acs.append(ac)
            result[name][oname] = [sn, sp, ac]
    return result        
        
def compare_annotation_files(fname1, fname2):        
    file1 = open(fname1)
    file2 = open(fname2)

    data1 = read_bed_transcripts(file1)
    data2 = read_bed_transcripts(file2)
       
    result = compare_annotation(data1, data2)
    for t2, x, exons in data2:
        if not result.has_key(t2) or result[t2] == {}:
            print "{0}\t{1}\t{2}\t{3}".format(exons[0][0], exons[0][1], exons[-1][2], 1)

    #print result 
    cluster = {}
    t2_cluster = {}
    for transcript2,d in result.items():
        for transcript1 in d.keys():
            if not cluster.has_key(transcript1):
                cluster[transcript1] = d.keys()
            else:
                for t in d.keys(): 
                    if not t in cluster[transcript1]:
                        cluster[transcript1].append(t) 
            for t1 in cluster[transcript1]:
                t2_cluster.setdefault(t1, []).append(transcript2)
    
    clusters = set([tuple(x) for x in cluster.values()])
    
    loc = {}
    for name,x,exons in data1:
        loc[name] = [exons[0][0], exons[0][1], exons[-1][2]]
    
    for cluster in clusters:
        t2s = set(t2_cluster[cluster[0]])
        min_acs = []
        for transcript1 in cluster:
            acs = []
            for transcript2 in t2s:
                if result[transcript2].has_key(transcript1):
                    sn, sp, ac = result[transcript2][transcript1]
                    #print transcript1, transcript2, sn, sp, ac
                    acs.append(1 - sp)
            min_acs.append(np.min(acs))
            #print transcript1, np.min(acs)
        #print "###"
        #print cluster, min_acs
        chrom = loc[cluster[0]][0]
        start = np.min([loc[t][1] for t in cluster])
        end = np.max([loc[t][2] for t in cluster])
        print "{0}\t{1}\t{2}\t{3}".format(chrom, start, end, np.mean(min_acs))
    #print loc   

        
        #for t in cluster:
    sys.exit()
    
    for k,v in result.items():
        if len(v.keys()) > 0:
            print "{0}\t{1}\t{2}".format(loc[k], k, 1 - np.max(v.values()))
        else:
            print "{0}\t{1}\t{2}".format(loc[k], k, 1.0)
        











fname1 = "/home/simon/prj/laevis/annotation/XENLA_JGI7b/gmap/xlaevisEST.gmap.bed"  
#fname2 = "/home/simon/prj/laevis/annotation/XENLA_JGI7b/gmap/XENLA_2013may.longest_cdna_annot.XENLA_JGIv7b.gmap.bed"
fname2 = "/home/simon/prj/laevis/pita/v1.32/pita_v1.32.bed"


#x = HTSeq.GenomicInterval( "chr3", 123203, 127245, "+" )
#sys.exit()




#!/usr/bin/env python
from rpy2.robjects.packages import importr
from rpy2.rinterface import RRuntimeError
import rpy2.robjects as robjects

# Try to load changepoint library, install it if it's not installed
try:
    cp = importr('changepoint')
except RRuntimeError:
    utils = importr('utils')
    # Choose first available mirror from the list
    utils.chooseCRANmirror(ind=1)
    utils.install_packages('changepoint')
    cp = importr('changepoint')

def cpt(l):
    """
    Call the mean.cpt function from the R changepoint package and 
    return the changepoint as a float.
    """
    
    vector = robjects.FloatVector([x for x in l])
    result = cp.cpt_mean(vector)

    try: 
        return float(cp.cpts(result)[0])
    except:
        return None



from pita.r_cpt import cpt
from pita.util import bed2exonbed
from pita.io import read_bed_transcripts
from tempfile import NamedTemporaryFile
import pybedtools
import subprocess as sp
import sys
import os
import numpy as np

def call_cpt(start, end, strand, data, min_reads=5, min_log2_ratio=1.5, upstream=False):
    """
    Determine UTR location from basepair-resolution coverage vector of reads.
    Return tuple of utr_start and utr_end if changepoint is found.
    """
    
    sys.stderr.write("{} {} {}\n".format(start, end, strand))
    pt_cutoff = 5
    counts = np.array(data)
    
    # Do calculations on reverse array if gene is on the - strand or
    # when predicting 5' UTR
    if (upstream and strand == "+") or strand == "-":
        counts = counts[::-1]
    
    pt = len(counts)
    ratio = 0
    while pt > pt_cutoff and ratio < 1:
        sys.stderr.write("cpt {}\n".format(pt))
        pt = cpt(counts[:pt])
        if pt > pt_cutoff:
            ratio = np.log2(counts[:pt].mean() / (counts[pt:].mean()) + 0.1)
        
    if pt > pt_cutoff:
        # Add to the changepoint while the number of reads is above min_reads
        while pt < len(counts) and counts[pt] >= min_reads:
            pt += 1
            
        m = counts[:pt].mean()
        if m >= min_reads and ratio >= min_log2_ratio:
            if (upstream and strand == "+") or strand == "-":
                utr_start = int(end) - pt
                utr_end = int(end)
            else:
                utr_start = int(start)
                utr_end = int(start) + pt
    
            return int(utr_start), int(utr_end)
           
def call_utr(inbed, bamfiles, utr5=False, utr3=True):
    """
    Call 3' UTR for all genes in a BED12 file based on RNA-seq reads 
    in BAM files.
    """
    
    # Load genes in BED file
    transcripts = read_bed_transcripts(open(inbed))
    
    # No genes
    if len(transcripts) == 0:
        return 

    td = dict([(t[0].split("|")[1] + "_", t[2]) for t in transcripts])
    
    #Trying to fix the scaffold struggles
    #td = {}
    #for t in transcripts:
    #    if "scaffold" not in t[0]:
    #        td[t[0].split("_")[1]+"_"] = t[2]
#	else:
#	    inter = t[0].split(":")
#            scafName = "_".join(inter[0].split("_")[2:4])
#	    pos = inter[1].split("_")[0]
#	    td[scafName+":"+pos+"_"] = t[2]

    # Create a BED6 file with exons, used to determine UTR boundaries 
    sys.stderr.write("Preparing temporary BED files\n")
    exonbed = NamedTemporaryFile(prefix="pita.", suffix=".bed")
    bed2exonbed(inbed, exonbed.name)

    # Determine boundaries using bedtools
    genes = pybedtools.BedTool(inbed)
    exons = pybedtools.BedTool(exonbed.name)
    
    tmp = NamedTemporaryFile(prefix="pita.", suffix=".bed")
    
    EXTEND = 10000
    sys.stderr.write("Determining gene boundaries determined by closest gene\n")
    for x in genes.closest(exons, D="a", io=True, iu=True):
        transcript = td[x[3]]
        
        # Extend to closest exon or EXTEND, whichever is closer
        extend = EXTEND
        if (int(x[-1]) >= 0) and (int(x[-1]) < extend):
            extend = int(x[-1])
        
        if transcript[0][-1] == "+":
            first = transcript[-1]
            first[2] += extend
        else:
            first = transcript[-0]
            first[1] -= extend
    
            if first[1] < 0:
                first[1] = 0
       
        tmp.write("{}\t{}\t{}\t{}\t{}\t{}\n".format(
            first[0],
            first[1],
            first[2],
            x[3],
            0,
            first[3]
            ))
        
    tmp.flush()
    
    tmpsam = NamedTemporaryFile(prefix="pita.", suffix=".sam")
    tmpbam = NamedTemporaryFile(prefix="pita.")
    
    # Retrieve header from first BAM file
    sp.call("samtools view -H {} > {}".format(bamfiles[0], tmpsam.name), shell=True)
    
    # Filter all BAM files for the specific regions. This runs much faster
    # then running bedtools coverage on all individual BAM files
    tmp_check = NamedTemporaryFile(prefix="pita.", suffix=".bam")
    cmd = "samtools view -L {} {} > {}"
    sys.stderr.write("Merging bam files\n")
    for bamfile in bamfiles:
        try:
            sp.check_call(cmd.format(tmp.name, bamfile, tmp_check.name), shell=True)
            sp.call("cat {} >> {}".format(tmp_check.name, tmpsam.name), shell=True)
        except sp.CalledProcessError as e:
            sys.stderr.write("Error in file {}, skipping:\n".format(bamfile))
            sys.stderr.write("{}\n".format(e))
    
    tmp_check.close()

    # Created sorted and index bam
<<<<<<< HEAD
    cmd = "samtools view -Sb {} | samtools sort -m 4G - {}"
=======
    cmd = "samtools view -Sb {} | samtools sort -m 10G - {}"
>>>>>>> master
    sp.call(cmd.format(tmpsam.name, tmpbam.name), shell=True)
    sp.call("samtools index {}.bam".format(tmpbam.name), shell=True)
    
    # Close and remove temporary SAM file
    tmpsam.close()
    
    sys.stderr.write("Calculating coverage\n")
    cmd = "bedtools coverage -abam {} -b {} -d -split "
    
    p = sp.Popen(cmd.format(tmpbam.name + ".bam", tmp.name), shell=True, stdout=sp.PIPE, bufsize=1)
    
    sys.stderr.write("Calling UTRs\n")
    
    data = []
    current = [None]
    utr = {}
    for line in iter(p.stdout.readline, b''):
        vals = line.strip().split("\t")
        if vals[3] != current[0]:
            if len(data) > 0:
                result = call_cpt(current[1], current[2], current[3], data, len(bamfile))
                #print result
                if result:
                    utr[current[0]] = result
            data = []
            current = [vals[3], int(vals[1]), int(vals[2]), vals[5]]
        data.append(int(vals[7]))
    if current[0]:
        result = call_cpt(current[1], current[2], current[3], data, len(bamfiles))
        if result:
            utr[current[0]] = result
    
    for fname in [tmpbam.name + ".bam", tmpbam.name + ".bam.bai"]:
        if os.path.exists(fname):
            os.unlink(fname)
    
    tmpbam.close()
    tmp.close()
    
    
    return utr

# A "nice" hack to implement 5' en 3' utr extension
def flip_bed_strands(bedfile):
    temp = NamedTemporaryFile(delete=False)
    for line in open(bedfile):
        line = line.strip().split("\t")
        if line[5] == "-":
            line[5] = "+"
        elif line[5] == "+":
            line[5] = "-"
        temp.write("\t".join(line)+"\n")
    temp.flush()
    return temp.name

def print_updated_bed(bedfile, bamfiles):
    #Extend the utr to the 5' end
    first = calculate_updated_bed(bedfile, bamfiles)

    #flipping the strands to extend the utr to the 3' end
    preSecond = flip_bed_strands(first)
    second = calculate_updated_bed(preSecond, bamfiles)

    #back to the correct strands
    final = flip_bed_strands(second)
    
    #print the utrExtended bed to the console
    for line in open(final):
        print(line.strip())


def calculate_updated_bed(bedfile, bamfiles):
    temp = NamedTemporaryFile(delete=False)

    utr = call_utr(bedfile, bamfiles)
    for line in open(bedfile):
        if line.startswith("track") or line[0] == "#":
            temp.write(line.strip())
            continue
        
        vals = line.strip().split("\t")
        start,end = int(vals[1]), int(vals[2])
        strand = vals[5]
        name = vals[3]
        thickstart, thickend = int(vals[6]), int(vals[7])
        exonsizes = [int(x) for x in vals[10].split(",") if x]
        exonstarts = [int(x) for x in vals[11].split(",") if x]
        
        if utr.has_key(name):
            sys.stderr.write("Updating {}\n".format(name))
            utr_start, utr_end = utr[name]
        
            if strand == "+":
                if utr_end < thickend:
                    sys.stderr.write("Setting end of {} to CDS end\n".format(name))
                    utr_end = thickend
                diff = exonsizes[-1] - (utr_end - utr_start)
                end -= diff
                
                exonsizes[-1] -= diff
    
                vals[2] = end
                vals[10] = ",".join([str(x) for x in exonsizes] + [""])
            else:
                if utr_start > thickstart:
                    sys.stderr.write("Setting start of {} to CDS start\n".format(name))
                    utr_start = thickstart
                diff = exonsizes[0] - (utr_end - utr_start)
                sys.stderr.write("{} {} {} diff: {}\n".format(utr_start, utr_end, exonsizes[0], diff))
                start += diff
                
                exonstarts = [0] + [x - diff for x in exonstarts[1:]]
                exonsizes[0] -= diff
    
                vals[1] = start
                vals[10] = ",".join([str(x) for x in exonsizes] + [""])
                vals[11] = ",".join([str(x) for x in exonstarts] + [""])
    
            temp.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(*vals))
        else:
            temp.write(line)
    temp.flush()
    return temp.name




from itertools import izip
import logging
import random
import re

import numpy as np
import networkx as nx 
from networkx.algorithms.connectivity import minimum_st_node_cut
from networkx.algorithms.flow import edmonds_karp

from pita.util import longest_orf,exons_to_seq

class DbCollection(object):
    def __init__(self, db, weights, prune=None, chrom=None):
        # dict with chrom as key
        self.logger = logging.getLogger("pita")

        self.db = db
        self.chrom = chrom

        # All transcript models will be stored as a directed (acyclic) graph
        self.graph = nx.DiGraph()

        # Store read counts of BAM files
        self.nreads = {}

        # Store extension used in BAM statistics
        self.extend = {}

        # store maximum weight per type
        self.max_id_value = {}
        for iw in weights:
            identifier = iw["name"]
            self.max_id_value[identifier] = 0
        
        # Load the exons
        self._load_exons(weights, chrom=chrom, prune=prune)
        
        # Load the introns
        self._load_splice_junctions(weights, chrom=chrom, prune=prune)

    def _load_exons(self, weights, chrom=None, prune=None):
        """
        Load exons from database in graph
        """
        self.logger.debug("Loading exons in graph")
        
        # Get filtering rules 
        if prune == None:
            prune = {}
        min_length = prune.get("exons", {}).get("min_length", None)
        max_length = prune.get("exons", {}).get("max_length", None)
        ev = prune.get("exons", {}).get("evidence", None)
        if min_length:
            self.logger.debug("Minimum length for exon: %s", 
                    min_length)
        if max_length:
            self.logger.debug("Maximum length for exon: %s", 
                    max_length)
        if ev:
            self.logger.debug("Minimum evidence for exon: %s", 
                    ev)
        
        for exon in self.db.get_exons(chrom, eager=True, 
                min_length=min_length, max_length=max_length, evidence=ev):
            self.add_feature(exon, weights)
    
    def _load_splice_junctions(self, weights, chrom=None, prune=None): 
        """
        Load splice junctions from database in graph
        """
        self.logger.debug("Loading introns in graph")
        
        # Get filtering rules 
        if prune == None:
            prune = {}
        min_reads = prune.get("introns", {}).get("min_reads", None)
        ev = prune.get("introns", {}).get("evidence", None)
        if min_reads:
            self.logger.debug("Minimum reads for splice junction: %s", 
                    min_reads)
        if ev:
            self.logger.debug("Minimum evidence for splice junction: %s", 
                    ev)
        
        # Load splice junctions    
        n = 0
        for junction in self.db.get_splice_junctions(chrom, 
                ev_count=ev, read_count=min_reads, eager=True):
            n += 1
            self.add_feature(junction, weights)
        self.logger.debug("%s introns were loaded", n)

    def add_feature(self, feature, weights):
        """ 
        Add feature to the graph
        """

        # Exon
        if feature.ftype == "exon":
            nodes = [
                    (feature.in_node(), "exon_in"), 
                    (feature.out_node(), "exon_out"),
                    ]
            for node,ftype in nodes:
                f = self.graph.add_node(node, ftype=ftype)
            self.graph.add_path([n[0] for n in nodes], 
                    weight=-1, ftype='exon')
            self._set_edge_weight(feature, nodes[0][0], nodes[1][0], weights)
        # Intron
        elif feature.ftype == "splice_junction":
            for e1,e2 in self.db.get_junction_exons(feature):
                if e1.strand == "-":
                    e1,e2 = e2,e1
                self.graph.add_path((e1.out_node(), e2.in_node()), 
                        ftype="splice_junction", weight=-1)
                self._set_edge_weight(feature, e1.out_node(), e2.in_node(), weights)

    def get_best_variants(self, weights):

        iweight = {}
        for iw in weights:
            identifier = iw["name"]
            weight = iw["weight"]
            iweight[identifier] = weight

        add = {}
        add_ends = []
        for i,model in enumerate(nx.weakly_connected_components(self.graph)):
            source = "source_{}".format(i + 1)
            sink = "sink_{}".format(i + 1)
            ends = [k for k,v in self.graph.out_degree(model).items() 
                    if self.graph.node[k].get('ftype', "") == "exon_out" and v == 0]
            for end in ends:
                add_ends.append((end, sink))

            for node in model:
                if self.graph.node[node].get('ftype', "") ==  "exon_in":
                    add.setdefault(source, []).append(node)

        for p in add_ends:
            self.graph.add_path(p, ftype="sink")

        for source, targets in add.items():
            for target in targets:
                self.graph.add_path((source, target), ftype='source')
                self._set_source_weight((source, target), weights)

        for n1,n2 in self.graph.edges():
            self.graph.edge[n1][n2]['weight'] = -0.01
            d = self.graph.edge[n1][n2]
            for k,v in self.max_id_value.items():
                if v > 0:
                    if k in d:
                        w = d[k] / float(v) * iweight[k]
                        self.graph.edge[n1][n2]['weight'] -= w 

            if self.graph.edge[n1][n2]['ftype'] == "exon":
                self.logger.debug("Edge: %s, %s", n1, n2)
                for k,v in d.items():
                    self.logger.debug("Key: %s, value: %s", k, v)

        for n1,n2 in self.graph.edges():
            d = self.graph.edge[n1][n2]

        for source, targets in add.items():
            sink = source.replace("source", "sink")
            try:
                pred,dis = nx.bellman_ford(self.graph, source)
                if not pred.has_key(sink):
                    continue
                t = sink
                best_variant = []
                while pred[t]:
                    best_variant.append(pred[t])
                    t = pred[t]

                p = re.compile(r'(.+):(\d+)([+-])')
                model = []
                strand = "+"
                for i in range(0, len(best_variant) - 1, 2):
                    n1,n2 = best_variant[i:i+2]
                    e = self._nodes_to_exon(n1, n2)
                    if e:
                        strand = e.strand
                        model.append(e)
                if strand == "+":
                    model = model[::-1]
                #print model
                if len(model) > 0:
                    yield model
    
            except Exception as e:
                raise
                self.logger.warning("Failed: %s", self.graph.edge[source].keys())
                self.logger.warning("%s", e)
    
    def _nodes_to_feature(self, n1, n2, feature): 
        
        p = re.compile(r'(.+):(\d+)([+-])')
        m = p.search(n1)
        chrom, start, strand = [m.group(x) for x in (1,2,3)]
        m = p.search(n2)
        start = int(start)
        end = int(m.group(2))
        if start > end:
            end,start = start,end
        e = self.db.fetch_feature((chrom, start, end, strand, feature, None))
        return e
    
    def _nodes_to_exon(self, n1, n2): 
        return self._nodes_to_feature(n1, n2, "exon")

    def _nodes_to_splice_junction(self, n1, n2): 
        return self._nodes_to_feature(n1, n2, "splice_junction")
    
    def _set_edge_weight(self, feature, n1, n2, weights):
        d = self.graph.edge[n1][n2]
        
        feature_stats = {}
        for f in feature.read_counts: 
            name = f.read_source.name
            feature_stats[name] = feature_stats.get(name, 0) + f.count
       
        for iw in weights:
            weight = iw["weight"]
            idtype = iw["type"]
            identifier = iw["name"]
            id_value = 0
            signal = feature_stats.get(identifier, 0)
            if d['ftype'] == "exon":
                length = feature.end - feature.start
                if idtype == "all":
                    id_value = signal
                elif idtype == "rpkm":
                    if signal > 0:
                        mreads = self.db.nreads(identifier) / 1e6
                        id_value = float(signal) / mreads / length * 1000.0
                elif idtype == "evidence":
                    id_value = len (feature.evidences)
                elif idtype == "length":
                    id_value = length
                elif idtype == "orf":
                    start, end = longest_orf(feature.seq)
                    id_value =  end - start
            elif d['ftype'] == "splice_junction":
                if idtype == "splice":
                    f = self._nodes_to_splice_junction(n1, n2)
                    id_value = feature_stats.get(identifier, 0)
            d[identifier] = id_value
            if id_value > self.max_id_value[identifier]:
                self.max_id_value[identifier] = id_value
        
    def _set_source_weight(self, edge, weights):
        n2 = edge[-1] 
        for iw in weights:
            weight = iw["weight"]
            idtype = iw["type"]
            identifier = iw["name"]
            if idtype == "first":
                for other in self.graph[n2]:
                    e = self._nodes_to_exon(n2, other)
                    signal = self.db.feature_stats(e, identifier)
                    self.graph.edge[edge[0]][n2][identifier] = signal
                    if signal > self.max_id_value[identifier]:
                        self.max_id_value[identifier] = signal

    def _model_to_path(self, model):
        """
        takes a list of Exons and returns a list of edges
        """
        
        # path is reversed if on minus strand
        if model[0].strand == "-":
            model = model[::-1]
        
        # add source node to the path
        source = [n for n in self.graph.predecessors(model[0].in_node())
                    if n.startswith("source")][0]
        path = [
                self.graph.edge[source][model[0].in_node()],
                self.graph.edge[model[0].in_node()][model[0].out_node()],
                    ]
        # add edge for every exon and intron
        for e1,e2 in zip(model[:-1], model[1:]):
            path += [
                    self.graph.edge[e1.out_node()][e2.in_node()],
                    self.graph.edge[e2.in_node()][e2.out_node()]
                    ]
        return path

    def get_weight(self, m):
        """
        return the total weight for a model
        """

        return sum([edge.get('weight', 0) for edge in self._model_to_path(m)])
    
    def filter_long(self, l=1000, evidence=2):
        """
        remove exon with length of at least <l> that is not supported by at least
        <evidence> sources
        """
        
        for exon in self.db.get_long_exons(self.chrom, l, evidence):
            out_edges = len(self.graph.out_edges([exon]))
            in_edges = len(self.graph.in_edges([exon]))
            self.logger.debug("Filter long: %s, in %s out %s", 
                    exon, in_edges, out_edges)

            if (in_edges >= 0 and out_edges >= 1 and exon.strand == "+" 
                    or in_edges >= 1 and out_edges >= 0 and exon.strand == "-"):
                self.logger.info("Removing long exon %s", exon)
                self.graph.remove_node(exon)




from pita.dbcollection import DbCollection
from pita.annotationdb import AnnotationDb
from pita.io import TabixIteratorAsFile, read_gff_transcripts, read_bed_transcripts
from pita.util import get_overlapping_models
import logging
import pysam
from pita.config import SEP

def load_chrom_data(conn, new, chrom, anno_files, data, index=None):
    logger = logging.getLogger("pita")
    
    try:
        # Read annotation files
        db = AnnotationDb(index=index, conn=conn, new=new)
        logger.debug("%s %s", chrom, id(db.session))
        logger.info("Reading annotation for %s", chrom)
        for name, fname, tabix_file, ftype, min_exons in anno_files:
            logger.info("Reading annotation from %s", fname)
            tabixfile = pysam.Tabixfile(tabix_file)
            #tabixfile = fname
            if chrom in tabixfile.contigs:
                fobj = TabixIteratorAsFile(tabixfile.fetch(chrom))
                if ftype == "bed":
                    it = read_bed_transcripts(fobj, fname, 
                            min_exons=min_exons, merge=10)
                elif ftype in ["gff", "gtf", "gff3"]:
                    it = read_gff_transcripts(fobj, fname, 
                            min_exons=min_exons, merge=10)
                for tname, source, exons in it:
                    db.add_transcript(
                            "{0}{1}{2}".format(name, SEP, tname), 
                            source, exons)
                del fobj    
            tabixfile.close()
            del tabixfile
        
        logger.info("Loading data for %s", chrom)

        for name, fname, span, extend in data:
            if span == "splice":
                logger.info("Reading splice data %s from %s", name, fname)
                db.get_splice_statistics(chrom, fname, name)
            else:
                logger.info("Reading BAM data %s from %s", name, fname)
                db.get_read_statistics(chrom, fname, name=name, span=span, extend=extend, nreads=None)
 
    except:
        logger.exception("Error on %s", chrom)
        raise

def get_chrom_models(conn, chrom, weight, repeats=None, prune=None, keep=None, filter_ev=None, experimental=None):
    if keep is None:
        keep = []
    if filter_ev is None:
        filter_ev = []
    if experimental is None:
        experimental = []

    logger = logging.getLogger("pita")
    logger.debug(str(weight)) 
    try:
        db = AnnotationDb(conn=conn)
        
        # Filter repeats
        if repeats:
            for x in repeats:
                db.filter_repeats(chrom, x)

        for ev in filter_ev:
            db.filter_evidence(chrom, ev, experimental) 
        
        mc = DbCollection(db, weight, prune=prune, chrom=chrom)
       
        # Remove short introns
        #mc.filter_short_introns()
      
        models = {}
        exons = {}
        logger.info("Calling transcripts for %s", chrom)
        for model in mc.get_best_variants(weight):
            genename = "{0}:{1}-{2}_".format(
                                        model[0].chrom,
                                        model[0].start,
                                        model[-1].end,
                                        )
                   
                
            logger.info("Best model: %s with %s exons", 
                    genename, len(model))
            models[genename] = [genename, model]
            
            for exon in model:
                exons[str(exon)] = [exon, genename]

        discard = {}
        if prune:
            logger.debug("Prune: {0}".format(prune))
            overlap = get_overlapping_models([x[0] for x in exons.values()])
            if len(overlap) > 1:
                logger.info("%s overlapping exons", len(overlap))
#                logger.warn("Overlap: {0}".format(overlap))
                
            gene_count = {}
            for e1, e2 in overlap:
                gene1 = exons[str(e1)][1]
                gene2 = exons[str(e2)][1]
                gene_count[gene1] = gene_count.setdefault(gene1, 0) + 1
                gene_count[gene2] = gene_count.setdefault(gene2, 0) + 1

            for e1, e2 in overlap:
                gene1 = exons[str(e1)][1]
                gene2 = exons[str(e2)][1]
                if not(gene1 in discard or gene2 in discard):
                    m1 = models[gene1][1]
                    m2 = models[gene2][1]
                
                    loc1,loc2 = sorted([m1, m2], cmp=lambda x,y: cmp(x[0].start, y[0].start))
                    l1 = float(loc1[-1].end - loc1[0].start)
                    l2 = float(loc2[-1].end - loc2[0].start)
                    if loc2[-1].end > loc1[-1].end:
                        overlap = float(loc1[-1].end - loc2[0].start)
                    else:
                        overlap = l2

                    #logger.info("Pruning {} vs. {}".format(str(m1),str(m2)))
                    #logger.info("1: {}, 2: {}, overlap: {}".format(
                    #    l1, l2, overlap))
                    #logger.info("Gene {} count {}, gene {} count {}".format(
                    #    str(gene1), gene_count[gene1], str(gene2), gene_count[gene2]
                    #    ))
#                   
                    prune_overlap = prune["overlap"]["fraction"]
                    if overlap / l1 < prune_overlap and overlap / l2 < prune_overlap:
                        logger.debug("Not pruning because fraction of overlap is too small!")
                        continue
                    
                    w1 = 0.0
                    w2 = 0.0
                    for d in prune["overlap"]["weights"]:
                        logger.debug("Pruning overlap: %s", d)
                        tmp_w1 = -mc.get_weight(m1)
                        tmp_w2 = -mc.get_weight(m2)
                        m = max((tmp_w1, tmp_w2))
                        if m > 0:
                            w1 += tmp_w1 / max((tmp_w1, tmp_w2))
                            w2 += tmp_w2 / max((tmp_w1, tmp_w2))

                    if w1 >= w2:
                        logger.info("Discarding %s", gene2)
                        discard[gene2] = 1
                    else:
                        logger.info("Discarding %s", gene1)
                        discard[gene1] = 1
        
        logger.info("Done calling transcripts for %s", chrom)
        result = [v for m,v in models.items() if not m in discard]
        #print "VV", result
        return [[name, [e.to_flat_exon() for e in exons]] for name, exons in result]

    except:
        logger.exception("Error on %s", chrom)
  
    return []



import logging
import yaml
import os
import sys
import pysam
from pita.io import _create_tabix

SAMTOOLS = "samtools"
TSS_FOUND = "v"
TSS_UPSTREAM = "u"
TSS_DOWNSTREAM = "a"
TSS_NOTFOUND = "x"
SEP = ":::"
VALID_TYPES = ["bed", "gff", "gff3", "gtf"]
DEBUG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]

class PitaConfig(object):
    def __init__(self):
        """ fname: name of yaml configuration file
        """
        self.logger = logging.getLogger("pita")
        
        self.maxentpath = ""

    def load(self, fname,  reannotate=False):
        # Parse YAML config file
        f = open(fname, "r")
        self.config = yaml.load(f)
        f.close()

        self.db_conn = "sqlite:///pita_database.db"
        if "database" in self.config:
            self.db_conn = self.config["database"]
        
        # Data directory
        self.base = "."
        if "data_path" in self.config:
            self.base = self.config["data_path"]

        # Prune overlaps
        self.prune = None
        if "prune" in self.config:
            self.prune = self.config["prune"]
        
        self.keep = []
        self.filter = []
        self.experimental = []
    
        # MaxEnt directory
        self.maxentpath = "" 
        if "maxent" in self.config:
            self.maxentpath = self.config["maxent"]
            
        # Pita UTR
        self.pitaUTR = False
        if "extendUtr" in self.config:
            self.pitaUTR = True

        # Scoring weight
        self.weight = {}
        if "scoring" in self.config:
            self.weight = self.config["scoring"]

        self._parse_repeats()
       
        # load annotation files
        self._parse_annotation(reannotate)
  
        # only use chromosome specified in config file
        self.chroms = self.chroms.keys()
        if "chromosomes" in self.config and self.config["chromosomes"]:
            if type(self.config["chromosomes"]) == type([]):
                self.chroms = self.config["chromosomes"]
            else:
                self.chroms = [self.config["chromosomes"]]

        # check the data files
        self._check_data_files()
        
        # output option
        self.min_protein_size = 20 
    
    def _parse_repeats(self):
        self.repeats = []
        if "repeats"in self.config:
            for d in self.config["repeats"]:
                fname = os.path.join(self.base, d["path"])
                tabix_fname = _create_tabix(fname, "bed")
                d["path"] = fname
                d["tabix"] = tabix_fname
 
                self.repeats.append(d)

    def _parse_annotation(self, reannotate=False):

        if not self.config.has_key("annotation") or len(self.config["annotation"]) == 0:
            self.logger.error("No annotation files specified.")
            sys.exit(1)
        
        self.anno_files = []
        self.chroms = {}
        for d in self.config["annotation"]:
            self.logger.debug("annotation: %s", d)
            fname = os.path.join(self.base, d["path"])
            t = d["type"].lower()
            min_exons = 2
            if d.has_key("min_exons"):
                min_exons = d["min_exons"]
            
            if d.has_key("keep") and d["keep"]:
                self.keep.append(fname)
           
            if d.has_key("filter") and d["filter"]:
                self.filter.append(fname)
            
            if d.has_key("experimental") and d["experimental"]:
                self.experimental.append(fname)

            if not t in VALID_TYPES:
                self.logger.error("Invalid type: %s", t)
                sys.exit(1)
            if not os.path.exists(fname):
                self.logger.error("File does not exist: %s", fname)
                sys.exit(1)
            else:
                tabix_file = ""
                if not reannotate:
                    tabix_file = _create_tabix(fname, t)
                    
                    # Save chromosome names
                    for chrom in pysam.Tabixfile(tabix_file).contigs:
                        self.chroms[chrom] = 1

                # Add file info
                self.anno_files.append([d["name"], fname, tabix_file, t, min_exons])
       
    def _check_data_files(self):
        # data  config
        self.logger.info("Checking data files")
        self.data = []
        if self.config.has_key("data") and self.config["data"]:
            for d in self.config["data"]:
                self.logger.debug("data: %s", d)
                d.setdefault("up", 0)
                d.setdefault("down", 0)
                if type("") == type(d["path"]):
                    d["path"] = [d["path"]]
       
                d.setdefault("feature", "all")
                if d["feature"] not in ["all", "start", "end", "splice"]:
                    self.logger.error("Incorrect span: %s", ["feature"])
                    sys.exit(1)

                fnames = [os.path.join(self.base, x) for x in d["path"]]
                for fname in fnames:
                    if not os.path.exists(fname):
                        self.logger.error("File does not exist: %s", fname)
                        sys.exit(1)
                  
                    if fname.endswith("bam") and not os.path.exists(fname + ".bai"):
                        self.logger.error("BAM file %s needs to be indexed!", fname)
                        sys.exit(1)

                    #if fname.endswith("bam"):
                    #    names_and_stats.append((fname, read_statistics(fname)))
                    #else:
                     #   names_and_stats.append((fname, None))
                row = [d["name"], fnames, d["feature"], (int(d["up"]), int(d["down"]))]
                self.data.append(row)

config = PitaConfig()



import logging
import sys
from pita.config import DEBUG_LEVELS

def setup_logging(basename, debug_level):
    debug_level = debug_level.upper()
    
    if not debug_level in DEBUG_LEVELS:
        sys.stderr.write("Invalid debug level {0}\n".format(debug_level))
        sys.stderr.write("Valid values are {0}\n".format(",".join(DEBUG_LEVELS)))
        sys.exit(1)

    logger = logging.getLogger("pita")
    logger.setLevel(getattr(logging, debug_level))
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(getattr(logging, debug_level))
    fh = logging.FileHandler("{0}.log".format(basename))
    fh.setFormatter(formatter)
    fh.setLevel(getattr(logging, debug_level))
    logger.addHandler(handler)
    logger.addHandler(fh)

    return logger




import os
import sys
import logging
from gimmemotifs.genome_index import GenomeIndex
from sqlalchemy import or_,and_,func
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session,sessionmaker,subqueryload
from pita.db_backend import Base,get_or_create,ReadSource,Feature,\
        FeatureReadCount,Evidence,FeatureEvidence 
from pita.util import read_statistics, get_splice_score
import yaml
from pita.io import exons_to_tabix_bed, tabix_overlap
from fluff.fluffio import get_binned_stats
from tempfile import NamedTemporaryFile

class AnnotationDb(object):
    def __init__(self, session=None, conn='mysql://pita:@localhost/pita', new=False, index=None):
        self.logger = logging.getLogger("pita")
        
        # initialize db session
        if session:
            self.session = session
        else:
            self._init_session(conn, new) 
        
        # index to retrieve sequence
        self.index = None
        if index:
            self.index = GenomeIndex(index)
   
        self.cache_splice_stats = {}
        self.cache_feature_stats = {}

    def _init_session(self, conn, new=False):
        self.engine = create_engine(conn)
        self.engine.raw_connection().connection.text_factory = str
        
        # recreate database
        if new:
            Base.metadata.drop_all(self.engine)
            Base.metadata.create_all(self.engine)
        
        Base.metadata.bind =self.engine
        Session = scoped_session(sessionmaker(bind=self.engine))
        self.session = Session()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.session.close()
    
    def dump_yaml(self):
        dump_dict = {}
        dump_dict['feature'] = [
                    [f.id, 
                    f.chrom.encode('ascii','ignore'), 
                    f.start, 
                    f.end, 
                    f.strand.encode('ascii','ignore'), 
                    f.ftype.encode('ascii','ignore'), 
                    f.seq.encode('ascii','ignore')] 
                for f in self.session.query(Feature)]
        
        dump_dict['read_source'] = [
                [r.id, 
                    r.name.encode('ascii','ignore'), 
                    r.source.encode('ascii','ignore'), 
                    r.nreads] 
                for r in self.session.query(ReadSource)]

        dump_dict['read_count'] = [
                [r.read_source_id, 
                    r.feature_id, 
                    r.count,r.span.encode('ascii','ignore'),
                    r.extend_up, 
                    r.extend_down]
                for r in self.session.query(FeatureReadCount)]
        
        dump_dict['evidence'] = [
                [r.id, r.name.encode('ascii','ignore'), 
                    r.source.encode('ascii','ignore')] 
                for r in self.session.query(Evidence)]
        
        dump_dict['feature_evidence'] = [
                [r.feature_id, r.evidence_id] 
                for r in self.session.query(FeatureEvidence)]

        return yaml.dump(dump_dict)
    
    def load_yaml(self, fname):
        data = yaml.load(open(fname))
        source_map = {}
        
        if not data['feature']:
            return 

        for old_id,name,fname,nreads in data['read_source']:
            r = get_or_create(self.session, ReadSource,
                    name=name, source=fname, nreads=nreads)
            self.session.commit()
            source_map[old_id] = r.id
    
    
        t = ["chrom","start","end","strand","ftype","seq"]
        self.engine.execute(
            Feature.__table__.insert(),
            [dict(zip(t, row[1:])) for row in data['feature']]
            )
         
        self.session.commit()
        
        first = self.fetch_feature(data['feature'][0][1:])
        last = self.fetch_feature(data['feature'][-1][1:])
    
        f_map = dict(zip([x[0] for x in data['feature']], range(first.id, last.id + 1)))
        data['read_count'] = [
                [source_map[row[0]]] + [f_map[row[1]]] + row[2:] for row in data['read_count']
            ]
        t = ["read_source_id", "feature_id", "count", "span", "extend_up", "extend_down"]
    
        self.engine.execute(
            FeatureReadCount.__table__.insert(),
            [dict(zip(t, row)) for row in data['read_count']]
            )
    
        if data['evidence']:
            t = ["name","source"]
            result = self.engine.execute(
                Evidence.__table__.insert(),
                [dict(zip(t, row[1:])) for row in data['evidence']]
                )
    
            self.session.commit()
            first = self.fetch_evidence(data['evidence'][0][1:])
            last = self.fetch_evidence(data['evidence'][-1][1:])
    
            ev_map = dict(zip([x[0] for x in data['evidence']], range(first.id, last.id + 1)))
    
            data['feature_evidence'] = [
                    [f_map[row[0]], ev_map[row[1]]] for row in data['feature_evidence']
                ]
    
            t = ["feature_id", "evidence_id"]
            self.engine.execute(
                FeatureEvidence.__table__.insert(),
                [dict(zip(t, row)) for row in data['feature_evidence']]
                )
    
    def add_transcript(self, name, source, exons):
        """
        Add a transcript to the database
        """

        # Sanity checks
        for e1, e2 in zip(exons[:-1], exons[1:]):
            if e1[0] != e2[0]:
                sys.stderr.write("{0} - {1}\n".format(e1, e2))
                raise ValueError("Different chromosomes!")
            if e2[1] <= e1[2]:
                sys.stderr.write("{0} - {1}\n".format(e1, e2))
                raise ValueError("exons overlap, or in wrong order")
            if e1[3] != e2[3]:
                sys.stderr.write("{0} - {1}\n".format(e1, e2))
                raise ValueError("strands don't match")
       
        chrom = exons[0][0]
        strand = exons[0][-1]
        
        evidence = get_or_create(self.session, Evidence,
                name = name,
                source=source)

        seqs = []
        for exon in exons:
            seq = ""
            real_seq = ""
            if self.index:
                seq = ""
                try:                    
                    seq = self.index.get_sequence(chrom, exon[1] - 20, exon[2] + 20, strand)
                    real_seq = seq[20:-20]
                except Exception:
                    real_seq = self.index.get_sequence(chrom, exon[1], exon[2], strand)
                seqs.append(seq)
            
            exon = get_or_create(self.session, Feature,
                             chrom = chrom,
                             start = exon[1],
                             end = exon[2],
                             strand = strand,
                             ftype = "exon",
                             seq = real_seq
                             ) 
            exon.evidences.append(evidence)

        splice_donors = []
        splice_acceptors = []
        for i,(start,end) in enumerate([(e1[2], e2[1]) for e1, e2 in zip(exons[0:-1], exons[1:])]):
            self.logger.debug("%s %s %s %s", chrom, start, end, strand)
            sj = get_or_create(self.session, Feature,
                             chrom = chrom,
                             start = start,
                             end = end,
                             strand = strand,
                             ftype = "splice_junction"
                             )
            sj.evidences.append(evidence)
            
            if strand == "+":
                if len(seqs) > (i + 1) and len(seqs[i]) > 46:
                    splice_donors.append(["{}_{}".format(name, i + 1), seqs[i][-23:-14]])
                if len(seqs) > (i + 2) and len(seqs[i + 1]) > 46:
                    f = ["{}_{}".format(name, i + 1), seqs[i + 1][:23]]
                    splice_acceptors.append(f)
            else:
                if len(seqs) > (i + 2) and len(seqs[i + 1]) > 46:
                    f = ["{}_{}".format(name, i + 1), seqs[i + 1][-23:-14]]
                    splice_donors.append(f)
                     
                if len(seqs) > (i + 1) and len(seqs[i]) > 46:
                    f = ["{}_{}".format(name, i + 1), seqs[i][:23]]
                    splice_acceptors.append(f)
        
        donor_score = get_splice_score(splice_donors, 5)
        acceptor_score = get_splice_score(splice_acceptors, 3)
        if donor_score + acceptor_score < 0:
            self.logger.warning("Skipping %s, splicing not OK!", name)
            self.session.rollback()
        else:
            self.session.commit()
    
    def get_features(self, ftype=None, chrom=None, eager=False, 
            min_length=None, max_length=None, evidence=0):
        
        query = self.session.query(Feature)
        query = query.filter(Feature.flag.op("IS NOT")(True))
            
        # pre-fetch associated read counts
        if eager:
            query = query.options(subqueryload('read_counts'))
        
        if chrom:
            query = query.filter(Feature.chrom == chrom)
        if ftype:
            query = query.filter(Feature.ftype == ftype)
        features = [f for f in query]
        
        # length filters
        if max_length:
            features = [f for f in features if
                    f.length <= max_length or len(f.evidences) >= evidence]

        if min_length:
            features = [f for f in features if
                    f.length >= min_length or len(f.evidences) >= evidence]
        
        return features

    def get_exons(self, chrom=None, eager=False, min_length=None, 
            max_length=None, evidence=0):
        
        return self.get_features(ftype="exon", chrom=chrom, eager=eager, 
                min_length=min_length, max_length=max_length, evidence=evidence)


    def get_splice_junctions(self, chrom=None, ev_count=None, read_count=None, max_reads=None, eager=False):
                
        features = []
        if ev_count and read_count:
            # All splices with no read, but more than one evidence source
            fs = self.session.query(Feature).\
                    filter(Feature.flag.op("IS NOT")(True)).\
                    filter(Feature.ftype == "splice_junction").\
                    filter(Feature.chrom == chrom).\
                    outerjoin(FeatureReadCount).\
                    group_by(Feature).\
                    having(func.sum(FeatureReadCount.count) < read_count)

            for splice in fs:
                self.logger.debug("Considering %s", splice)        
                for evidence in splice.evidences:
                    self.logger.debug(str(evidence))
                if len(splice.evidences) >= ev_count:
                    features.append(splice)
                else:
                    self.logger.debug("not enough evidence for {}".format(splice))
            
            fs = self.session.query(Feature).\
                    filter(Feature.flag.op("IS NOT")(True)).\
                    filter(Feature.ftype == "splice_junction").\
                    filter(Feature.chrom == chrom).\
                    outerjoin(FeatureReadCount).\
                    group_by(Feature).\
                    having(func.sum(FeatureReadCount.count) == None)

            for splice in fs:
                self.logger.debug("Considering %s (no reads)", splice)        
                for evidence in splice.evidences:
                    self.logger.debug(str(evidence))
                if len(splice.evidences) >= ev_count:
                    features.append(splice)
                else:
                    self.logger.debug("not enough evidence for {}".format(splice))

            # All splices with more than x reads
            fs = self.session.query(Feature).\
                    filter(Feature.flag.op("IS NOT")(True)).\
                    filter(Feature.ftype == "splice_junction").\
                    filter(Feature.chrom == chrom).\
                    outerjoin(FeatureReadCount).\
                    group_by(Feature).\
                    having(func.sum(FeatureReadCount.count) >= read_count)
            for f in fs:
                self.logger.debug("Considering %s (reads)", f)        
                features.append(f)
            #features += [f for f in fs]
        
        elif max_reads:
            fs = self.session.query(Feature).\
                    filter(Feature.flag.op("IS NOT")(True)).\
                    filter(Feature.ftype == "splice_junction").\
                    filter(Feature.chrom == chrom).\
                    outerjoin(FeatureReadCount).\
                    group_by(Feature).\
                    having(func.sum(FeatureReadCount.count) == None)

            features += [f for f in fs if len(f.evidences) > 0]
            fs = self.session.query(Feature).\
                    filter(Feature.flag.op("IS NOT")(True)).\
                    filter(Feature.ftype == "splice_junction").\
                    filter(Feature.chrom == chrom).\
                    outerjoin(FeatureReadCount).\
                    group_by(Feature).\
                    having(func.sum(FeatureReadCount.count) < max_reads)
            features += [f for f in fs if len(f.evidences) > 0]  
        else:
            features = self.get_features(ftype="splice_junction", chrom=chrom, eager=eager)
        return features 

    def get_longest_3prime_exon(self, chrom, start5, strand):
        if strand == "+":
        
            q = self.session.query(Feature).\
                filter(Feature.flag.op("IS NOT")(True)).\
                filter(Feature.ftype == "exon").\
                filter(Feature.chrom == chrom).\
                filter(Feature.strand == strand).\
                filter(Feature.start == start5).\
                order_by(Feature.end)
            return q.all()[-1]
        else:
            q = self.session.query(Feature).\
                filter(Feature.ftype == "exon").\
                filter(Feature.chrom == chrom).\
                filter(Feature.strand == strand).\
                filter(Feature.end == start5).\
                order_by(Feature.end)
            return q.all()[0]

    
    def get_long_exons(self, chrom, l, evidence):
        query = self.session.query(Feature)
        query = query.filter(Feature.flag.op("IS NOT")(True))
        query = query.filter(Feature.ftype == 'exon')
        query = query.filter(Feature.chrom == chrom)
        query = query.filter(Feature.end - Feature.start >= l)
        return [e for e in query if len(e.evidences) <= evidence]
    
    def filter_repeats(self, chrom, rep):
        """ Flag all exons that overlap with a specified fraction
        with a repeat track
        """

        self.logger.warn("Filtering repeats: %s with fraction %s", 
                os.path.basename(rep["path"]), rep["fraction"])
        
        exons = self.get_features("exon", chrom)
        exon_tabix = exons_to_tabix_bed(exons) 
        
        overlap_it = tabix_overlap(exon_tabix, rep["tabix"], chrom, rep["fraction"])
        exon_ids = [int(iv[3]) for iv in overlap_it]
        
        chunk = 20
        for i in range(0, len(exon_ids), chunk):
            self.logger.warn("Filtering %s", exon_ids[i:i + chunk])
            self.session.query(Feature).\
                    filter(Feature.id.in_(exon_ids[i:i + chunk])).\
                    update({Feature.flag:True}, synchronize_session=False)
            self.session.commit()
            

        #fobj = TabixIteratorAsFile(tabixfile.fetch(chrom))
        #for line in fobj:
        #    print line

    def filter_evidence(self, chrom, source, experimental):
        self.logger.debug("Filtering %s", source)
        #query = self.session.query(Feature).\
        #        update({Feature.flag:False}, synchronize_session=False)
        #self.session.commit()

        # Select all features that are supported by other evidence
        n = self.session.query(Feature.id).\
                        join(FeatureEvidence).\
                        join(Evidence).\
                        filter(Evidence.source != source).\
                        filter(Evidence.source not in experimental).\
                        filter(Feature.chrom == chrom).\
                        subquery("n")
       
        # Select the total number of transcript from this source 
        # per feature
        s = self.session.query(Feature.id, func.count('*').label('total')).\
               join(FeatureEvidence).\
               join(Evidence).\
               filter(Evidence.source == source).\
               filter(Feature.chrom == chrom).\
               group_by(Feature.id).\
               subquery("s")

        # Select all features where support from this source
        # is only 1 transcript and which is not supported by any 
        # other sources
        a = self.session.query(Feature.id).filter(and_(
            Feature.id == s.c.id,
            s.c.total == 1)).\
            filter(Feature.id.notin_(n)).\
            subquery("a")

        #ids = [i[0] for i in query]
        
        # Flag features
        self.session.query(Feature).\
                filter(Feature.id.in_(a)).\
                update({Feature.flag:True}, synchronize_session=False)
        self.session.commit()
        
    def get_read_statistics(self, chrom, fnames, name, span="all", extend=(0,0), nreads=None):

        if span not in ["all", "start", "end"]:
            raise Exception("Incorrect span: {}".format(span))
        
        tmp = NamedTemporaryFile(delete=False)
        estore = {}
        self.logger.debug("Writing exons to file %s", tmp.name)
        exons =  self.get_exons(chrom)
        if len(exons) == 0:
            return
        
        for exon in exons:
            start = exon.start
            end = exon.end
            if span == "start":
                if exon.strand == "+":
                    end = start
                elif exon.strand == "-":
                    start = end
            if span == "end":
                if exon.strand == "+":
                    start = end
                elif exon.strand == "-":
                    end = start
            
            if exon.strand == "-":
                start -= extend[1]
                end += extend[0]
            else:
                start -= extend[0]
                end += extend[1]
            if start < 0:
                start = 0

            estr = "{}:{}-{}".format(exon.chrom, start, end)

            if estr in estore:
                estore[estr].append(exon)
            else:
                estore[estr] = [exon]
                tmp.write("{}\t{}\t{}\t{}\t{}\t{}\n".format(
                    exon.chrom,
                    start,
                    end,
                    str(exon),
                    0,
                    exon.strand
                ))
        tmp.flush()

        if type("") == type(fnames):
            fnames = [fnames]

        for i, fname in enumerate(fnames):
            self.logger.debug("Creating read_source for %s %s", name, fname)
            read_source = get_or_create(self.session, ReadSource, name=name, source=fname)
            self.session.commit() 
            #rmrepeats = False
            if fname.endswith("bam") and (not nreads or not nreads[i]):
                #rmrepeats = True
                self.logger.debug("Counting reads in %s", fname)
                read_source.nreads = read_statistics(fname)

            self.logger.debug("Getting overlap from %s", fname)
            result = get_binned_stats(tmp.name, fname, 1, rpkm=False, rmdup=False, rmrepeats=False)

            self.logger.debug("Reading results, save to exon stats")

            insert_vals = []
            for row in result:
                try:
                    vals = row.strip().split("\t")
                    e = "%s:%s-%s" % (vals[0], vals[1], vals[2])
                    c = float(vals[3])
                    for exon in estore[e]:
                        insert_vals.append([read_source.id, exon.id, c, span, extend[0], extend[1]])
                except:
                    self.logger.info("binned_stat line skipped: {}".format(row))
            t =  ["read_source_id", "feature_id", "count", "span", "extend_up", "extend_down"]
            result = self.engine.execute(
                    FeatureReadCount.__table__.insert(),
                    [dict(zip(t,row)) for row in insert_vals]
                    )
                
        tmp.close()

    def get_splice_statistics(self, chrom, fnames, name):
        if type("") == type(fnames):
            fnames = [fnames]

        for fname in fnames:
            self.logger.debug("Getting splicing data from %s", fname)
            read_source = get_or_create(self.session, 
                    ReadSource, name=name, source=fname)
            self.session.commit()
            for line in open(fname):
                vals = line.strip().split("\t")
                if vals[0] == chrom:
                    start, end, c = [int(x) for x in vals[1:4]]
                    strand = vals[5]
                    
                    splice = get_or_create(self.session, Feature,
                             chrom = chrom,
                             start = start,
                             end = end,
                             strand = strand,
                             ftype = "splice_junction"
                             ) 
                    self.session.commit()

                    count = get_or_create(self.session, FeatureReadCount,
                            feature_id = splice.id,
                            read_source_id = read_source.id)
                
                    if not count.count:
                        count.count = c
                    else:
                        count.count += c
            
                    self.session.commit()    
    
    def get_junction_exons(self, junction):
        
        left = self.session.query(Feature).filter(and_(
            Feature.chrom == junction.chrom,
            Feature.strand == junction.strand,
            Feature.end == junction.start,
            Feature.ftype == "exon"
            )).\
            filter(Feature.flag.op("IS NOT")(True))

        
        right = self.session.query(Feature).filter(and_(
            Feature.chrom == junction.chrom,
            Feature.strand == junction.strand,
            Feature.start == junction.end,
            Feature.ftype == "exon"
            )).\
            filter(Feature.flag.op("IS NOT")(True))

        exon_pairs = []
        for e1 in left:
            for e2 in right:
                exon_pairs.append((e1, e2))
        return exon_pairs

    def clear_stats_cache(self):
        self.cache_feature_stats = {}
        self.cache_splice_stats = {}
    
    def feature_stats(self, feature, identifier):
        if "{}{}".format(feature, identifier) not in self.cache_feature_stats:
            q = self.session.query(FeatureReadCount, ReadSource).join(ReadSource)
            q = q.filter(FeatureReadCount.feature_id == feature.id)
            q = q.filter(ReadSource.name == identifier)
            self.cache_feature_stats["{}{}".format(feature, identifier)] = sum([x[0].count for x in q.all()])

        return self.cache_feature_stats["{}{}".format(feature, identifier)]

    def splice_stats(self, exon1, exon2, identifier):
        if "{}{}{}".format(self, exon1, exon2) not in self.cache_splice_stats:
            q = self.session.query(Feature)
            q = q.filter(Feature.ftype == "splice_junction")
            q = q.filter(Feature.chrom == exon1.chrom)
            q = q.filter(Feature.strand == exon1.strand)
            q = q.filter(Feature.start == exon1.end)
            q = q.filter(Feature.end == exon2.start)

            splice = q.first()
        
            self.cache_splice_stats["{}{}{}".format(self, exon1, exon2)] = self.feature_stats(splice, identifier)

        return self.cache_splice_stats["{}{}{}".format(self, exon1, exon2)]

    def intron_splice_stats(self, intron, identifier):
        if "{}{}".format(self, intron) not in self.cache_splice_stats:
            q = self.session.query(Feature)
            q = q.filter(Feature.ftype == "splice_junction")
            q = q.filter(Feature.chrom == intron.chrom)
            q = q.filter(Feature.strand == intron.strand)
            q = q.filter(Feature.start == intron.start)
            q = q.filter(Feature.end == intron.end)

            splice = q.first()
        
            self.cache_splice_stats["{}{}".format(self, intron)] = self.feature_stats(splice, identifier)

        return self.cache_splice_stats["{}{}".format(self, intron)]
    
    def nreads(self, identifier):
        q = self.session.query(ReadSource)
        q = q.filter(ReadSource.name == identifier)

        return sum([s.nreads for s in q.all()])
    
    def get_splice_count(self, e1, e2):
        counts = self.session.query(func.sum(FeatureReadCount.count)).\
                join(Feature).\
                filter(Feature.chrom == e1.chrom).\
                filter(Feature.start == e1.end).\
                filter(Feature.end == e2.start).\
                filter(Feature.ftype == "splice_junction").\
                group_by(Feature.id).all()
        return sum([int(x[0]) for x in counts])
    
    def fetch_feature(self, f):
        """ Feature as list """
        chrom, start, end, strand, ftype, seq = f

        feature = self.session.query(Feature).\
                filter(Feature.chrom == chrom).\
                filter(Feature.start == start).\
                filter(Feature.end == end).\
                filter(Feature.strand == strand).\
                filter(Feature.ftype == ftype)
                
        if seq:
            feature = feature.filter(Feature.seq == seq)
        result = feature.first()
        return result
    
    def fetch_evidence(self, f):
        """ Feature as list """
        name, source = f
        evidence = self.session.query(Evidence).\
            filter(Evidence.name == name).\
            filter(Evidence.source == source)

        result = evidence.first()
        return result

    def get_transcript_statistics(self, exons):
        stats = []
        for exon in exons:
            q = self.session.query(ReadSource, 
                    func.sum(FeatureReadCount.count)).\
                    join(FeatureReadCount).\
                    join(Feature).\
                    filter(Feature.chrom == exon[0]).\
                    filter(Feature.start == exon[1]).\
                    filter(Feature.end == exon[2]).\
                    filter(Feature.strand == exon[3]).\
                    group_by(ReadSource.name)
            stats.append(dict([(row[0].name, int(row[1])) for row in q.all()]))
        return stats

                   



from BCBio import GFF
import sys
import os
import logging 
from tempfile import NamedTemporaryFile
import subprocess as sp
import pysam
import pybedtools

def _create_tabix(fname, ftype):
    logger = logging.getLogger("pita")
    tabix_file = ""
    logger.info("Creating tabix index for %s", os.path.basename(fname))
    logger.debug("Preparing %s for tabix", fname)
    tmp = NamedTemporaryFile(prefix="pita", delete=False)
    preset = "gff"
    if ftype == "bed":
        cmd = "sort -k1,1 -k2g,2 {0} | grep -v track | grep -v \"^#\" > {1}"
        preset = "bed"
    elif ftype in ["gff", "gff3", "gtf"]:
        cmd = "sort -k1,1 -k4g,4 {0} | grep -v \"^#\" > {1}"

    # Sort the input file
    logger.debug(cmd.format(fname, tmp.name))
    sp.call(cmd.format(fname, tmp.name), shell=True)
    # Compress using bgzip
    logger.debug("compressing %s", tmp.name)
    tabix_file = tmp.name + ".gz"
    pysam.tabix_compress(tmp.name, tabix_file)
    tmp.close()
    # Index (using tabix command line, as pysam.index results in a Segmentation fault
    logger.debug("indexing %s", tabix_file)
    sp.call("tabix {0} -p {1}".format(tabix_file, preset), shell=True)
    return tabix_file

def exons_to_tabix_bed(exons):
    logger = logging.getLogger("pita")
    logger.debug("Converting %s exons to tabix bed", len(exons))
    tmp = NamedTemporaryFile(prefix="pita", delete=False)
    logger.debug("Temp name %s", tmp.name)
    for exon in sorted(exons, cmp=lambda x,y: cmp([x.chrom, x.start], [y.chrom, y.start])):
        tmp.write("{}\t{}\t{}\t{}\t{}\t{}\n".format(
            exon.chrom, exon.start, exon.end, exon.id, "0", exon.strand))

    tmp.close()
    tabix_fname = _create_tabix(tmp.name, "bed")
    return tabix_fname

def tabix_overlap(fname1, fname2, chrom, fraction):
    logger = logging.getLogger("pita")
    logger.debug("TABIX overlap between %s and %s, %s", fname1, fname2, fraction)

    tab1 = pysam.Tabixfile(fname1)
    tab2 = pysam.Tabixfile(fname2)

    if not ((chrom in tab1.contigs) and (chrom in tab2.contigs)):
        return
    
    fobj1 = TabixIteratorAsFile(tab1.fetch(chrom))
    tmp1 = NamedTemporaryFile(prefix="pita.", delete=False)
    for line in fobj1.readlines():
        tmp1.write("{}\n".format(line.strip()))
    
    fobj2 = TabixIteratorAsFile(tab2.fetch(chrom))
    tmp2 = NamedTemporaryFile(prefix="pita.", delete=False)
    for line in fobj2.readlines():
        tmp2.write("{}\n".format(line.strip()))

    tmp1.flush()
    tmp2.flush()

    b1 = pybedtools.BedTool(tmp1.name)
    b2 = pybedtools.BedTool(tmp2.name)

    intersect = b1.intersect(b2, f=fraction)

    tmp1.close()
    tmp2.close()
    
    for f in intersect:
        yield f

def merge_exons(starts, sizes, l=0):
    merge = []
    for i, (start1, start2, size) in enumerate(zip(starts[:-1], starts[1:], sizes[:-1])):
        if start1 + size + l >= start2:
            merge.append(i + 1)
    
    if len(merge) == 0:
        return starts, sizes
   
    i = len(starts) - 1
    new_starts = []
    new_sizes = []
    
    while i >= 0:
        if i in merge:
            j = i - 1
            while j in merge:
                j -= 1
            new_starts.append(starts[j])
            new_sizes.append(starts[i] - starts[j] + sizes[i])
            
            i = j - 1
        else:
            new_starts.append(starts[i])
            new_sizes.append(sizes[i])
            i -= 1
    
    return new_starts[::-1], new_sizes[::-1]

def _gff_type_iterator(feature, ftypes):
    if feature.type in ftypes:
        yield feature
    else:
        for feature in feature.sub_features:
            for f in _gff_type_iterator(feature, ftypes):
                yield f

def read_gff_transcripts(fobj, fname="", min_exons=1, merge=0):
    
    # Setup logging
    logger = logging.getLogger('pita')
  
    if merge > 0:
        logger.warning("Merging exons not yet implemented for GFF files!")

    #limits = dict(gff_type = ["mRNA", "exon"])
    smap = {"1":"+",1:"+","-1":"-",-1:"-", None:"+"}
    transcripts = []
    for rec in GFF.parse(fobj):
        chrom = rec.id
        for feature in rec.features:
            #logger.debug("feature: {0}", feature)
            
            for gene in _gff_type_iterator(feature, ['mRNA', 'transcript', 'inferred_parent']):
                #logger.debug("Adding gene: {0}", gene)
                exons = []
                #logger.debug("subfeatures: {0}", gene.sub_features)
                for exon in [f for f in gene.sub_features if f.type == 'exon']:
                    #link[gene.id] = link.setdefault(gene.id, 0) + 1
                    start = int(exon.location.start.position)# - 1    
                    end = int(exon.location.end.position)
                    strand = smap[exon.strand]
                    exons.append([chrom, start, end, strand])
                logger.debug("%s: %s - %s exons", fname, gene.id, len(exons))
                if len(exons) >= min_exons:
                    transcripts.append([gene.id, fname, exons])

    return transcripts

def read_bed_transcripts(fobj, fname="", min_exons=1, merge=0):
    
    # Setup logging
    logger = logging.getLogger('pita')
    
    names = {}
    transcripts = []
    line = fobj.readline()
    while line:
        if not line.startswith("track"):
            #logger.debug(line)
            try:
                vals = line.strip().split("\t")
                
                i = 1
                name = "%s|%s|%s" % (vals[0], vals[3], i)
                while names.has_key(name):
                    i += 1
                    name = "%s|%s|%s" % (vals[0], vals[3], i)
                names[name] = 1
                
                chromStart = int(vals[1])
                
                sizes = [int(x) for x in vals[10].strip(",").split(",")]
                starts = [int(x) for x in vals[11].strip(",").split(",")] 
                    
                starts, sizes = merge_exons(starts, sizes, l=merge)
                
                exons = [[vals[0], 
                          chromStart + start, 
                          chromStart + start + size, 
                          vals[5]]
                          for start, size in zip(starts, sizes)
                         ]
                    
                if len(exons) >= min_exons:
                    logger.debug("read_bed: adding %s", vals[3])
                    transcripts.append([name, fname, exons])
                else:
                    logger.debug("read_bed: not adding %s, filter on minimum exons", vals[3])
            
            except:
                print "Error parsing BED file"
                print line
                raise
                sys.exit(1)
        
        line = fobj.readline()
    return transcripts

class TabixIteratorAsFile:
    def __init__(self, x):
        self.x = x

    def read(self):
        return None

    def readline(self):
        try:
            return self.x.next()
        except StopIteration:
            return None

    def readlines(self):
        line = self.readline()
        while line:
            yield line
            line = self.readline()







from pita.config import SAMTOOLS
from pita.config import config
from subprocess import Popen, PIPE
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
import subprocess as sp
from tempfile import NamedTemporaryFile
import logging

logger = logging.getLogger('pita')

def read_statistics(fname, rmrepeat=False, rmdup=False, mapped=False):
    """ Count number of reads in BAM file.
    Optional arguments rmrepeat and rmdup do nothing for now
    """

    if rmrepeat:
        pass
    if rmdup:
        pass

    cmd = "{0} idxstats {1} | awk '{{total += $3 + $4}} END {{print total}}'"

    p = Popen(
             cmd.format(SAMTOOLS, fname), 
             shell=True, 
             stdout=PIPE, 
             stderr=PIPE
             )

    stdout,_ = p.communicate()

    n = int(stdout.strip())

    return n

def get_overlapping_models(exons):
    overlap = []
    sorted_exons = sorted(exons, cmp=lambda x,y: cmp(x.start, y.start))

    for i, exon in enumerate(sorted_exons):
        j = i + 1
        while j < len(sorted_exons) and sorted_exons[j].start <= exon.end:
            if exon.overlap(sorted_exons[j]):
                overlap.append([exon, sorted_exons[j]])
            j += 1
    
    return overlap

def exons_to_seq(exons):
    seq = ""
    if exons[0].strand == "-":
        exons = exons[::-1]

    for e in exons:
        if e.seq:
            seq = "".join((seq, e.seq))
        else:
            logger.error("exon %s has no sequence", e)

    return seq

def longest_orf(seq, do_prot=False):
    if isinstance(seq, list):
        seq = exons_to_seq(seq)

    dna = Seq(seq, IUPAC.ambiguous_dna)

    my_cmp = lambda x,y: cmp(len(x), len(y))

    orfs = []
    prots = []
    for i in range(3):
        seq = dna[i:]

        # BioPython doesn't like translating DNA with a length that's not
        # a multiple of 3
        if len(seq) % 3:
            seq = seq[:-(len(seq) % 3)]

        prot = str(seq.translate())
        #putative_orfs = [re.sub(r'^[^M]*', "", o) for o in prot.split("*")]
        putative_orfs = [o for o in prot.split("*")]
        longest = sorted(putative_orfs, cmp=my_cmp)[-1]
        prots.append(longest)
        start = prot.find(longest) * 3 + i
        end = start + (len(longest) + 1) * 3
        orfs.append((start,end))

    if do_prot:
        return sorted(prots, cmp=lambda x,y: cmp(len(x), len(y)))[-1]
    else:
        return sorted(orfs, cmp=lambda x,y: cmp(x[1] - x[0], y[1] - y[0]))[-1]

def find_genomic_pos(pos, exons):
    if exons[0].strand == "-":
        for e in exons[::-1]:
            size = e.end - e.start
            if pos < size:
                return e.end - pos
            else:
                pos -= size
        return exons[0].start
    else:
        for e in exons:
            size = e.end - e.start
            if pos < size:
                return e.start + pos
            else:
                pos -= size
        return exons[-1].end

def to_genomic_orf(start, end, exons):
    genomic_start = find_genomic_pos(start, exons)
    genomic_end = find_genomic_pos(end,  exons)
    if exons[0].strand == "-":
        return genomic_end, genomic_start
    else:
        return genomic_start, genomic_end

def model_to_bed(exons, genename=None):
    line_format = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}"
    
    chrom = exons[0].chrom
    chromStart = exons[0].start
    chromEnd = exons[-1].end
    
    if not genename:
        genename = "{0}:{1}-{2}".format(chrom, chromStart, chromEnd)
    
    thickStart, thickEnd = chromStart, chromEnd

    if exons[0].seq:
        orf_start, orf_end = longest_orf(exons)
        thickStart, thickEnd = to_genomic_orf(orf_start, orf_end, exons)

    sizes = ",".join([str(exon.end - exon.start) for exon in exons]) + ","
    starts = ",".join([str(exon.start - chromStart) for exon in exons]) + ","

    return line_format.format(
                            chrom,
                            chromStart,
                            chromEnd,
                            genename,
                            600,
                            exons[0].strand,
                            thickStart,
                            thickEnd,
                            "0,0,0",
                            len(exons),
                            sizes,
                            starts
                            )



def get_splice_score(a, s_type=5):
    if s_type not in [3,5]:
        raise Exception("Invalid splice type {}, should be 3 or 5".format(s_type))
    
    maxent = config.maxentpath
    if not maxent:
        raise Exception("Please provide path to the score5.pl and score3.pl maxent scripts in config file")

    tmp = NamedTemporaryFile()
    for name,seq in a:
        tmp.write(">{}\n{}\n".format(name,seq))
    tmp.flush()
    cmd = "perl score{}.pl {}".format(s_type, tmp.name)
    p = sp.Popen(cmd, shell=True, cwd=maxent, stdout=sp.PIPE)
    score = 0
    for line in p.stdout.readlines():
        vals = line.strip().split("\t")
        if len(vals) > 1:
            try:
                score += float(vals[-1])
            except ValueError:
                logger.error("valueError, skipping: {}".format(vals))
            except:
                logger.error("Something unexpected happened")
    return score

def bed2exonbed(inbed, outbed):
    with open(outbed, "w") as out:
        for line in open(inbed):
            if line.startswith("#") or line.startswith("track"):
                out.write(line)
            else:
                vals = line.strip().split("\t")
        
                exonsizes = [int(x) for x in vals[10].split(",") if x]
                exonstarts = [int(x) for x in vals[11].split(",") if x]
                for exon_start, exon_size in zip(exonstarts, exonsizes):
                    out.write("{}\t{}\t{}\t{}\t{}\t{}\n".format(
                        vals[0], 
                        int(vals[1]) + exon_start, 
                        int(vals[1]) + exon_start + exon_size, 
                        vals[3], 
                        vals[4], 
                        vals[5]))



from warnings import warn
from sqlalchemy import Column, ForeignKey, Integer, String, Text, UniqueConstraint, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship,sessionmaker,scoped_session
from sqlalchemy import and_, event
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method

from pita.exon import Exon


Base = declarative_base()

class FeatureEvidence(Base):
    __tablename__ = 'feature_evidence'
    feature_id = Column(Integer, ForeignKey('feature.id'), primary_key=True)
    evidence_id = Column(Integer, ForeignKey('evidence.id'), primary_key=True)
    evidence = relationship("Evidence")
    feature = relationship("Feature")

class Feature(Base):
    __tablename__ = 'feature'
    __table_args__ = (
             UniqueConstraint(
                'chrom', 
                'start', 
                'end', 
                'strand', 
                'ftype',
                name='uix_1'),
            )
    id = Column(Integer, primary_key=True)
    chrom = Column(String(250), nullable=False) 
    start = Column(Integer, nullable=False) 
    end = Column(Integer, nullable=False)
    strand = Column(String(1), nullable=False)
    ftype = Column(String(250), nullable=False) 
    seq = Column(Text(), default="") 
    flag = Column(Boolean(), default=False)
    _evidences = relationship('FeatureEvidence')
    evidences = association_proxy('_evidences', 'evidence',
                    creator=lambda _i: FeatureEvidence(evidence=_i),
                )
    read_counts = relationship('FeatureReadCount')
   
    @hybrid_property
    def length(self):
        return self.end - self.start

    def __str__(self):
        return self.to_loc()
    
    def to_loc(self):
        return "{}:{}{}{}".format(
                self.chrom, self.start, self.strand, self.end
                )
    
    def to_sloc(self):
        return "{}{}{}".format(
                self.start, self.strand, self.end
                )
    
    def in_node(self):
        if self.strand == "+":
            return "{}:{}{}_{}".format(
                    self.chrom, self.start, self.strand, self.ftype + "_in")
        else: 
            return "{}:{}{}_{}".format(
                    self.chrom, self.end, self.strand, self.ftype + "_in")
    
    def out_node(self):
        if self.strand == "-":
            return "{}:{}{}_{}".format(
                    self.chrom, self.start, self.strand, self.ftype + "_out")
        else: 
            return "{}:{}{}_{}".format(
                    self.chrom, self.end, self.strand, self.ftype + "_out")
  

        
    def to_flat_exon(self):
        e = Exon(self.chrom, self.start, self.end, self.strand)
        e.seq = self.seq
        return e

    def overlap(self, exon, strand=True, fraction=False):
        if fraction:
            warn("the 'fraction' argument in overlap is not yet implemented")

        if strand and self.strand != exon.strand:
            return 0
        if exon.start >= self.start and exon.start <= self.end:
            return self.end - exon.start
        if exon.end >= self.start and exon.end <= self.end:
            return exon.end - self.start
        return 0

class Evidence(Base):
    __tablename__ = "evidence"
    id = Column(Integer, primary_key=True)
    name = Column(String(250))
    source = Column(String(250))

class ReadSource(Base):
    __tablename__ = "read_source"
    id = Column(Integer, primary_key=True)
    name = Column(String(250))
    source = Column(String(250))
    nreads = Column(Integer, default=0)

class FeatureReadCount(Base):
    __tablename__ = "read_count"
    read_source_id = Column(Integer, ForeignKey('read_source.id'), primary_key=True)
    feature_id = Column(Integer, ForeignKey('feature.id'), primary_key=True)
    read_source = relationship("ReadSource")
    feature = relationship("Feature")
    count = Column(Integer, default=0)
    span = Column(String(50), default="all", primary_key=True)
    extend_up = Column(Integer, default=0, primary_key=True)
    extend_down = Column(Integer, default=0, primary_key=True)

def get_or_create(session, model, **kwargs):
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        instance = model(**kwargs)
        session.add(instance)
        return instance



from distutils.core import setup
from setuptools.command.test import test as TestCommand
import sys

VERSION = "1.76"
DESCRIPTION = """
pita - pita improves transcript annotation
"""
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup (name = 'pita',
        version = VERSION,
        description = DESCRIPTION,
        author='Simon van Heeringen',
        author_email='s.vanheeringen@ncmls.ru.nl',
        license='MIT',
        url = 'https://github.com/simonvh/pita',
        download_url = 'https://github.com/simonvh/pita/tarball/'+VERSION,
        packages=[
            'pita'
        ],
        scripts=[
            "scripts/pita",
            "scripts/pita_utr",
            "scripts/bed12togff3",
            "scripts/gff3tobed12",
            "scripts/flatbread",
            "scripts/breadcrumb",
        ],
        data_files=[],
        tests_require=['pytest'],
      #  dependency_links = [
       #     'https://github.com/simonvh/gimmemotifs/archive/0.8.7.tar.gz#egg=gimmemotifs-0.8.7',
      #      ],
        install_requires=[
                        "numpy",
                        "SQLAlchemy",
                        "gimmemotifs > 0.8.6",
                        "pysam >= 0.9",
                        "pyyaml",
                        "HTSeq",
                        "bcbio-gff",
                        "biopython",
                        "networkx >= 1.10",
                        ],
        cmdclass = {'test': PyTest},
)
